var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __metadata = (this && this.__metadata) || function (k, v) {
    if (typeof Reflect === "object" && typeof Reflect.metadata === "function") return Reflect.metadata(k, v);
};
var __param = (this && this.__param) || function (paramIndex, decorator) {
    return function (target, key) { decorator(target, key, paramIndex); }
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import { expect } from "chai";
import { all, DI, inject, lazy, optional, Registration, singleton, } from "./di.js";
describe("DI.get", function () {
    let container;
    // eslint-disable-next-line mocha/no-hooks
    beforeEach(function () {
        container = DI.createContainer();
    });
    describe("@lazy", function () {
        class Bar {
        }
        let Foo = class Foo {
            constructor(provider) {
                this.provider = provider;
            }
        };
        Foo = __decorate([
            __param(0, lazy(Bar)),
            __metadata("design:paramtypes", [Function])
        ], Foo);
        it("@singleton", function () {
            const bar0 = container.get(Foo).provider();
            const bar1 = container.get(Foo).provider();
            expect(bar0).to.equal(bar1);
        });
        it("@transient", function () {
            container.register(Registration.transient(Bar, Bar));
            const bar0 = container.get(Foo).provider();
            const bar1 = container.get(Foo).provider();
            expect(bar0).to.not.equal(bar1);
        });
    });
    describe("@scoped", function () {
        describe("true", function () {
            let ScopedFoo = class ScopedFoo {
            };
            ScopedFoo = __decorate([
                singleton({ scoped: true })
            ], ScopedFoo);
            describe("Foo", function () {
                const constructor = ScopedFoo;
                it("children", function () {
                    const root = DI.createContainer();
                    const child1 = root.createChild();
                    const child2 = root.createChild();
                    const a = child1.get(constructor);
                    const b = child2.get(constructor);
                    const c = child1.get(constructor);
                    expect(a).to.equal(c, "a and c are the same");
                    expect(a).to.not.equal(b, "a and b are not the same");
                    expect(root.has(constructor, false)).to.equal(false, "root has class");
                    expect(child1.has(constructor, false)).to.equal(true, "child1 has class");
                    expect(child2.has(constructor, false)).to.equal(true, "child2 has class");
                });
                it("root", function () {
                    const root = DI.createContainer();
                    const child1 = root.createChild();
                    const child2 = root.createChild();
                    const a = root.get(constructor);
                    const b = child2.get(constructor);
                    const c = child1.get(constructor);
                    expect(a).to.equal(c, "a and c are the same");
                    expect(a).to.equal(b, "a and b are the same");
                    expect(root.has(constructor, false)).to.equal(true, "root has class");
                    expect(child1.has(constructor, false)).to.equal(false, "child1 does not have class");
                    expect(child2.has(constructor, false)).to.equal(false, "child2 does not have class");
                });
            });
        });
        describe("false", function () {
            let ScopedFoo = class ScopedFoo {
            };
            ScopedFoo = __decorate([
                singleton({ scoped: false })
            ], ScopedFoo);
            describe("Foo", function () {
                const constructor = ScopedFoo;
                it("children", function () {
                    const root = DI.createContainer();
                    const child1 = root.createChild();
                    const child2 = root.createChild();
                    const a = child1.get(constructor);
                    const b = child2.get(constructor);
                    const c = child1.get(constructor);
                    expect(a).to.equal(c, "a and c are the same");
                    expect(a).to.equal(b, "a and b are the same");
                    expect(root.has(constructor, false)).to.equal(true, "root has class");
                    expect(child1.has(constructor, false)).to.equal(false, "child1 has class");
                    expect(child2.has(constructor, false)).to.equal(false, "child2 has class");
                });
            });
            describe("default", function () {
                let DefaultFoo = class DefaultFoo {
                };
                DefaultFoo = __decorate([
                    singleton
                ], DefaultFoo);
                const constructor = DefaultFoo;
                it("children", function () {
                    const root = DI.createContainer();
                    const child1 = root.createChild();
                    const child2 = root.createChild();
                    const a = child1.get(constructor);
                    const b = child2.get(constructor);
                    const c = child1.get(constructor);
                    expect(a).to.equal(c, "a and c are the same");
                    expect(a).to.equal(b, "a and b are the same");
                    expect(root.has(constructor, false)).to.equal(true, "root has class");
                    expect(child1.has(constructor, false)).to.equal(false, "child1 has class");
                    expect(child2.has(constructor, false)).to.equal(false, "child2 has class");
                });
            });
        });
    });
    describe("@optional", function () {
        it("with default", function () {
            let Foo = class Foo {
                constructor(test = "hello") {
                    this.test = test;
                }
            };
            Foo = __decorate([
                __param(0, optional("key")),
                __metadata("design:paramtypes", [String])
            ], Foo);
            expect(container.get(Foo).test).to.equal("hello");
        });
        it("no default, but param allows undefined", function () {
            let Foo = class Foo {
                constructor(test) {
                    this.test = test;
                }
            };
            Foo = __decorate([
                __param(0, optional("key")),
                __metadata("design:paramtypes", [String])
            ], Foo);
            expect(container.get(Foo).test).to.equal(undefined);
        });
        it("no default, param does not allow undefind", function () {
            let Foo = class Foo {
                constructor(test) {
                    this.test = test;
                }
            };
            Foo = __decorate([
                __param(0, optional("key")),
                __metadata("design:paramtypes", [String])
            ], Foo);
            expect(container.get(Foo).test).to.equal(undefined);
        });
        it("interface with default", function () {
            const Strings = DI.createContext(x => x.instance([]));
            let Foo = class Foo {
                constructor(test) {
                    this.test = test;
                }
            };
            Foo = __decorate([
                __param(0, optional(Strings)),
                __metadata("design:paramtypes", [Array])
            ], Foo);
            expect(container.get(Foo).test).to.equal(undefined);
        });
        it("interface with default and default in constructor", function () {
            const MyStr = DI.createContext(x => x.instance("hello"));
            let Foo = class Foo {
                constructor(test = "test") {
                    this.test = test;
                }
            };
            Foo = __decorate([
                __param(0, optional(MyStr)),
                __metadata("design:paramtypes", [String])
            ], Foo);
            expect(container.get(Foo).test).to.equal("test");
        });
        it("interface with default registered and default in constructor", function () {
            const MyStr = DI.createContext(x => x.instance("hello"));
            container.register(MyStr);
            let Foo = class Foo {
                constructor(test = "test") {
                    this.test = test;
                }
            };
            Foo = __decorate([
                __param(0, optional(MyStr)),
                __metadata("design:paramtypes", [String])
            ], Foo);
            expect(container.get(Foo).test).to.equal("hello");
        });
    });
    describe("intrinsic", function () {
        describe("bad", function () {
            it("Array", function () {
                let Foo = class Foo {
                    constructor(test) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    singleton,
                    __param(0, inject(Array)),
                    __metadata("design:paramtypes", [Array])
                ], Foo);
                expect(() => container.get(Foo)).throws();
            });
            it("ArrayBuffer", function () {
                let Foo = class Foo {
                    constructor(test) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    singleton,
                    __param(0, inject(ArrayBuffer)),
                    __metadata("design:paramtypes", [ArrayBuffer])
                ], Foo);
                expect(() => container.get(Foo)).throws();
            });
            it("Boolean", function () {
                let Foo = class Foo {
                    // eslint-disable-next-line @typescript-eslint/ban-types
                    constructor(test) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    singleton,
                    __param(0, inject(Boolean)),
                    __metadata("design:paramtypes", [Boolean])
                ], Foo);
                expect(() => container.get(Foo)).throws();
            });
            it("DataView", function () {
                let Foo = class Foo {
                    constructor(test) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    singleton,
                    __param(0, inject(DataView)),
                    __metadata("design:paramtypes", [DataView])
                ], Foo);
                expect(() => container.get(Foo)).throws();
            });
            it("Date", function () {
                let Foo = class Foo {
                    constructor(test) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    singleton,
                    __param(0, inject(Date)),
                    __metadata("design:paramtypes", [Date])
                ], Foo);
                expect(() => container.get(Foo)).throws();
            });
            it("Error", function () {
                let Foo = class Foo {
                    constructor(test) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    singleton,
                    __param(0, inject(Error)),
                    __metadata("design:paramtypes", [Error])
                ], Foo);
                expect(() => container.get(Foo)).throws();
            });
            it("EvalError", function () {
                let Foo = class Foo {
                    constructor(test) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    singleton,
                    __param(0, inject(EvalError)),
                    __metadata("design:paramtypes", [EvalError])
                ], Foo);
                expect(() => container.get(Foo)).throws();
            });
            it("Float32Array", function () {
                let Foo = class Foo {
                    constructor(test) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    singleton,
                    __param(0, inject(Float32Array)),
                    __metadata("design:paramtypes", [Float32Array])
                ], Foo);
                expect(() => container.get(Foo)).throws();
            });
            it("Float64Array", function () {
                let Foo = class Foo {
                    constructor(test) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    singleton,
                    __param(0, inject(Float64Array)),
                    __metadata("design:paramtypes", [Float64Array])
                ], Foo);
                expect(() => container.get(Foo)).throws();
            });
            it("Function", function () {
                let Foo = class Foo {
                    // eslint-disable-next-line @typescript-eslint/ban-types
                    constructor(test) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    singleton,
                    __param(0, inject(Function)),
                    __metadata("design:paramtypes", [Function])
                ], Foo);
                expect(() => container.get(Foo)).throws();
            });
            it("Int8Array", function () {
                let Foo = class Foo {
                    constructor(test) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    singleton,
                    __param(0, inject(Int8Array)),
                    __metadata("design:paramtypes", [Int8Array])
                ], Foo);
                expect(() => container.get(Foo)).throws();
            });
            it("Int16Array", function () {
                let Foo = class Foo {
                    constructor(test) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    singleton,
                    __param(0, inject(Int16Array)),
                    __metadata("design:paramtypes", [Int16Array])
                ], Foo);
                expect(() => container.get(Foo)).throws();
            });
            it("Int32Array", function () {
                let Foo = class Foo {
                    constructor(test) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    singleton,
                    __param(0, inject(Int32Array)),
                    __metadata("design:paramtypes", [Int16Array])
                ], Foo);
                expect(() => container.get(Foo)).throws();
            });
            it("Map", function () {
                let Foo = class Foo {
                    constructor(test) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    singleton,
                    __param(0, inject(Map)),
                    __metadata("design:paramtypes", [Map])
                ], Foo);
                expect(() => container.get(Foo)).throws();
            });
            it("Number", function () {
                let Foo = class Foo {
                    // eslint-disable-next-line @typescript-eslint/ban-types
                    constructor(test) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    singleton,
                    __param(0, inject(Number)),
                    __metadata("design:paramtypes", [Number])
                ], Foo);
                expect(() => container.get(Foo)).throws();
            });
            it("Object", function () {
                let Foo = class Foo {
                    // eslint-disable-next-line @typescript-eslint/ban-types
                    constructor(test) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    singleton,
                    __param(0, inject(Object)),
                    __metadata("design:paramtypes", [Object])
                ], Foo);
                expect(() => container.get(Foo)).throws();
            });
            it("Promise", function () {
                let Foo = class Foo {
                    constructor(test) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    singleton,
                    __param(0, inject(Promise)),
                    __metadata("design:paramtypes", [Promise])
                ], Foo);
                expect(() => container.get(Foo)).throws();
            });
            it("RangeError", function () {
                let Foo = class Foo {
                    constructor(test) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    singleton,
                    __param(0, inject(RangeError)),
                    __metadata("design:paramtypes", [RangeError])
                ], Foo);
                expect(() => container.get(Foo)).throws();
            });
            it("ReferenceError", function () {
                let Foo = class Foo {
                    constructor(test) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    singleton,
                    __param(0, inject(ReferenceError)),
                    __metadata("design:paramtypes", [ReferenceError])
                ], Foo);
                expect(() => container.get(Foo)).throws();
            });
            it("RegExp", function () {
                let Foo = class Foo {
                    constructor(test) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    singleton,
                    __param(0, inject(RegExp)),
                    __metadata("design:paramtypes", [RegExp])
                ], Foo);
                expect(() => container.get(Foo)).throws();
            });
            it("Set", function () {
                let Foo = class Foo {
                    constructor(test) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    singleton,
                    __param(0, inject(Set)),
                    __metadata("design:paramtypes", [Set])
                ], Foo);
                expect(() => container.get(Foo)).throws();
            });
            // if (typeof SharedArrayBuffer !== 'undefined') {
            //   it('SharedArrayBuffer', function () {
            //     @singleton
            //     class Foo {
            //       public constructor(private readonly test: SharedArrayBuffer) {
            //       }
            //     }
            //     assert.throws(() => container.get(Foo));
            //   });
            // }
            it("String", function () {
                let Foo = class Foo {
                    // eslint-disable-next-line @typescript-eslint/ban-types
                    constructor(test) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    singleton,
                    __param(0, inject(String)),
                    __metadata("design:paramtypes", [String])
                ], Foo);
                expect(() => container.get(Foo)).throws();
            });
            it("SyntaxError", function () {
                let Foo = class Foo {
                    constructor(test) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    singleton,
                    __param(0, inject(SyntaxError)),
                    __metadata("design:paramtypes", [SyntaxError])
                ], Foo);
                expect(() => container.get(Foo)).throws();
            });
            it("TypeError", function () {
                let Foo = class Foo {
                    constructor(test) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    singleton,
                    __param(0, inject(TypeError)),
                    __metadata("design:paramtypes", [TypeError])
                ], Foo);
                expect(() => container.get(Foo)).throws();
            });
            it("Uint8Array", function () {
                let Foo = class Foo {
                    constructor(test) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    singleton,
                    __param(0, inject(Uint8Array)),
                    __metadata("design:paramtypes", [Uint8Array])
                ], Foo);
                expect(() => container.get(Foo)).throws();
            });
            it("Uint8ClampedArray", function () {
                let Foo = class Foo {
                    constructor(test) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    singleton,
                    __param(0, inject(Uint8ClampedArray)),
                    __metadata("design:paramtypes", [Uint8ClampedArray])
                ], Foo);
                expect(() => container.get(Foo)).throws();
            });
            it("Uint16Array", function () {
                let Foo = class Foo {
                    constructor(test) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    singleton,
                    __param(0, inject(Uint16Array)),
                    __metadata("design:paramtypes", [Uint16Array])
                ], Foo);
                expect(() => container.get(Foo)).throws();
            });
            it("Uint32Array", function () {
                let Foo = class Foo {
                    constructor(test) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    singleton,
                    __param(0, inject(Uint32Array)),
                    __metadata("design:paramtypes", [Uint32Array])
                ], Foo);
                expect(() => container.get(Foo)).throws();
            });
            it("UriError", function () {
                let Foo = class Foo {
                    constructor(test) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    singleton,
                    __param(0, inject(URIError)),
                    __metadata("design:paramtypes", [URIError])
                ], Foo);
                expect(() => container.get(Foo)).throws();
            });
            it("WeakMap", function () {
                let Foo = class Foo {
                    constructor(test) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    singleton,
                    __param(0, inject(WeakMap)),
                    __metadata("design:paramtypes", [WeakMap])
                ], Foo);
                expect(() => container.get(Foo)).throws();
            });
            it("WeakSet", function () {
                let Foo = class Foo {
                    constructor(test) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    singleton,
                    __param(0, inject(WeakSet)),
                    __metadata("design:paramtypes", [WeakSet])
                ], Foo);
                expect(() => container.get(Foo)).throws();
            });
        });
        describe("good", function () {
            it("@all()", function () {
                let Foo = class Foo {
                    constructor(test) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    __param(0, all("test")),
                    __metadata("design:paramtypes", [Array])
                ], Foo);
                expect(container.get(Foo).test).to.eql([]);
            });
            it("@optional()", function () {
                let Foo = class Foo {
                    constructor(test = null) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    __param(0, optional("test")),
                    __metadata("design:paramtypes", [Object])
                ], Foo);
                expect(container.get(Foo).test).to.equal(null);
            });
            it("undef instance, with constructor default", function () {
                container.register(Registration.instance("test", undefined));
                let Foo = class Foo {
                    constructor(test = []) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    __param(0, inject("test")),
                    __metadata("design:paramtypes", [Array])
                ], Foo);
                expect(container.get(Foo).test).to.eql([]);
            });
            it("can inject if registered", function () {
                container.register(Registration.instance(String, "test"));
                let Foo = class Foo {
                    constructor(test) {
                        this.test = test;
                    }
                };
                Foo = __decorate([
                    singleton,
                    __param(0, inject(String)),
                    __metadata("design:paramtypes", [String])
                ], Foo);
                expect(container.get(Foo).test).to.equal("test");
            });
        });
    });
});
describe("DI.getAsync", () => {
    it("calls the registration locator for unknown keys", () => __awaiter(void 0, void 0, void 0, function* () {
        const key = "key";
        const instance = {};
        const asyncRegistrationLocator = (key) => __awaiter(void 0, void 0, void 0, function* () {
            return Registration.instance(key, instance);
        });
        const container = DI.createContainer({
            asyncRegistrationLocator
        });
        const found = yield container.getAsync(key);
        expect(found).equals(instance);
        const foundAgain = container.get(key);
        expect(foundAgain).equals(instance);
    }));
    it("calls the registration locator for unknown dependencies", () => __awaiter(void 0, void 0, void 0, function* () {
        const key1 = "key";
        const instance1 = {};
        const key2 = "key2";
        const instance2 = {};
        const key3 = "key3";
        const instance3 = {};
        const asyncRegistrationLocator = (key) => __awaiter(void 0, void 0, void 0, function* () {
            switch (key) {
                case key1:
                    return Registration.instance(key1, instance1);
                case key2:
                    return Registration.instance(key2, instance2);
                case key3:
                    return Registration.instance(key3, instance3);
            }
            throw new Error();
        });
        const container = DI.createContainer({
            asyncRegistrationLocator
        });
        let Test = class Test {
            constructor(one, two, three) {
                this.one = one;
                this.two = two;
                this.three = three;
            }
        };
        Test = __decorate([
            __param(0, inject(key1)),
            __param(1, inject(key2)),
            __param(2, inject(key3)),
            __metadata("design:paramtypes", [Object, Object, Object])
        ], Test);
        container.register(Registration.singleton(Test, Test));
        const found = yield container.getAsync(Test);
        expect(found.one).equals(instance1);
        expect(found.two).equals(instance2);
        expect(found.three).equals(instance3);
        const foundAgain = container.get(Test);
        expect(foundAgain).equals(found);
    }));
    it("calls the registration locator for a hierarchy of unknowns", () => __awaiter(void 0, void 0, void 0, function* () {
        const key1 = "key";
        const instance1 = {};
        const key2 = "key2";
        const instance2 = {};
        const key3 = "key3";
        const instance3 = {};
        let Test = class Test {
            constructor(one, two, three) {
                this.one = one;
                this.two = two;
                this.three = three;
            }
        };
        Test = __decorate([
            __param(0, inject(key1)),
            __param(1, inject(key2)),
            __param(2, inject(key3)),
            __metadata("design:paramtypes", [Object, Object, Object])
        ], Test);
        let Test2 = class Test2 {
            constructor(test) {
                this.test = test;
            }
        };
        Test2 = __decorate([
            __param(0, inject(Test)),
            __metadata("design:paramtypes", [Test])
        ], Test2);
        const asyncRegistrationLocator = (key) => __awaiter(void 0, void 0, void 0, function* () {
            switch (key) {
                case key1:
                    return Registration.instance(key1, instance1);
                case key2:
                    return Registration.instance(key2, instance2);
                case key3:
                    return Registration.instance(key3, instance3);
                case Test:
                    return Registration.singleton(key, Test);
                case Test2:
                    return Registration.transient(key, Test2);
            }
            throw new Error();
        });
        const container = DI.createContainer({
            asyncRegistrationLocator
        });
        const found = yield container.getAsync(Test2);
        expect(found.test.one).equals(instance1);
        expect(found.test.two).equals(instance2);
        expect(found.test.three).equals(instance3);
        const foundTest = container.get(Test);
        expect(foundTest).equals(found.test);
        const foundTransient = container.get(Test2);
        expect(foundTransient).not.equals(found);
        expect(foundTransient).instanceOf(Test2);
        expect(foundTransient.test.one).equals(instance1);
        expect(foundTransient.test.two).equals(instance2);
        expect(foundTransient.test.three).equals(instance3);
        expect(foundTransient.test).equals(foundTest);
    }));
});
