var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __metadata = (this && this.__metadata) || function (k, v) {
    if (typeof Reflect === "object" && typeof Reflect.metadata === "function") return Reflect.metadata(k, v);
};
import { expect } from "chai";
import { Metadata } from "./metadata.js";
import { emptyArray } from "./platform.js";
function decorator() { return (target) => target; }
describe("Metadata", () => {
    describe(`getDesignParamTypes()`, () => {
        it(`returns emptyArray if the class has no constructor or decorators`, () => {
            class Foo {
            }
            const actual = Metadata.getDesignParamTypes(Foo);
            expect(actual).equal(emptyArray);
        });
        it(`returns emptyArray if the class has a decorator but no constructor`, () => {
            let Foo = class Foo {
            };
            Foo = __decorate([
                decorator()
            ], Foo);
            const actual = Metadata.getDesignParamTypes(Foo);
            expect(actual).equal(emptyArray);
        });
        it(`returns emptyArray if the class has no constructor args or decorators`, () => {
            class Foo {
                constructor() { return; }
            }
            const actual = Metadata.getDesignParamTypes(Foo);
            expect(actual).equal(emptyArray);
        });
        it(`returns emptyArray if the class has constructor args but no decorators`, () => {
            class Bar {
            }
            class Foo {
                constructor(bar) {
                    this.bar = bar;
                }
            }
            const actual = Metadata.getDesignParamTypes(Foo);
            expect(actual).equal(emptyArray);
        });
        it(`returns emptyArray if the class has constructor args and the decorator is applied via a function call`, () => {
            class Bar {
            }
            class Foo {
                constructor(bar) {
                    this.bar = bar;
                }
            }
            decorator()(Foo);
            const actual = Metadata.getDesignParamTypes(Foo);
            expect(actual).equal(emptyArray);
        });
        it(`returns an empty mutable array if the class has a decorator but no constructor args`, () => {
            let Foo = class Foo {
                constructor() { return; }
            };
            Foo = __decorate([
                decorator(),
                __metadata("design:paramtypes", [])
            ], Foo);
            const actual = Metadata.getDesignParamTypes(Foo);
            expect(actual).not.equal(emptyArray);
            expect(actual).length(0);
        });
        describe(`falls back to Object for declarations that cannot be statically analyzed`, () => {
            for (const argCtor of [
                class Bar {
                },
                function () { return; },
                () => { return; },
                class {
                },
                {},
                Error,
                Array,
                (class Bar {
                }).prototype,
                (class Bar {
                }).prototype.constructor
            ]) {
                let FooDecoratorInvocation = class FooDecoratorInvocation {
                    constructor(arg) {
                        this.arg = arg;
                    }
                };
                FooDecoratorInvocation = __decorate([
                    decorator(),
                    __metadata("design:paramtypes", [Object])
                ], FooDecoratorInvocation);
                it(`FooDecoratorInvocation { constructor(${argCtor.name}) }`, () => {
                    const actual = Metadata.getDesignParamTypes(FooDecoratorInvocation);
                    expect(actual).length(1);
                    expect(actual[0]).equal(Object);
                });
                let FooDecoratorNonInvocation = class FooDecoratorNonInvocation {
                    constructor(arg) {
                        this.arg = arg;
                    }
                };
                FooDecoratorNonInvocation = __decorate([
                    decorator,
                    __metadata("design:paramtypes", [Object])
                ], FooDecoratorNonInvocation);
                it(`FooDecoratorNonInvocation { constructor(${argCtor.name}) }`, () => {
                    const actual = Metadata.getDesignParamTypes(FooDecoratorInvocation);
                    expect(actual).length(1);
                    expect(actual[0]).equal(Object);
                });
            }
        });
        describe(`returns the correct types for valid declarations`, () => {
            class Bar {
            }
            class Foo {
            }
            class Baz {
            }
            describe(`decorator invocation`, () => {
                it(`Class { constructor(public arg: Bar) }`, () => {
                    let FooBar = class FooBar {
                        constructor(arg) {
                            this.arg = arg;
                        }
                    };
                    FooBar = __decorate([
                        decorator(),
                        __metadata("design:paramtypes", [Bar])
                    ], FooBar);
                    const actual = Metadata.getDesignParamTypes(FooBar);
                    expect(actual).length(1);
                    expect(actual[0]).equal(Bar);
                });
                it(`Class { constructor(public arg1: Bar, public arg2: Foo) }`, () => {
                    let FooBar = class FooBar {
                        constructor(arg1, arg2) {
                            this.arg1 = arg1;
                            this.arg2 = arg2;
                        }
                    };
                    FooBar = __decorate([
                        decorator(),
                        __metadata("design:paramtypes", [Bar, Foo])
                    ], FooBar);
                    const actual = Metadata.getDesignParamTypes(FooBar);
                    expect(actual).length(2);
                    expect(actual[0]).equal(Bar);
                    expect(actual[1]).equal(Foo);
                });
                it(`Class { constructor(public arg1: Bar, public arg2: Foo, public arg3: Baz) }`, () => {
                    let FooBar = class FooBar {
                        constructor(arg1, arg2, arg3) {
                            this.arg1 = arg1;
                            this.arg2 = arg2;
                            this.arg3 = arg3;
                        }
                    };
                    FooBar = __decorate([
                        decorator(),
                        __metadata("design:paramtypes", [Bar, Foo, Baz])
                    ], FooBar);
                    const actual = Metadata.getDesignParamTypes(FooBar);
                    expect(actual).length(3);
                    expect(actual[0]).equal(Bar);
                    expect(actual[1]).equal(Foo);
                    expect(actual[2]).equal(Baz);
                });
            });
        });
    });
    describe(`getAnnotationParamTypes()`, () => {
        it("returns emptyArray if the class has no annotations", () => {
            class Foo {
            }
            const actual = Metadata.getAnnotationParamTypes(Foo);
            expect(actual).equal(emptyArray);
        });
        it("returns added annotations", () => {
            class Foo {
            }
            const a = Metadata.getOrCreateAnnotationParamTypes(Foo);
            a.push("test");
            const actual = Metadata.getAnnotationParamTypes(Foo);
            expect(actual).length(1);
            expect(actual[0]).equal("test");
        });
    });
    describe(`getOrCreateAnnotationParamTypes()`, () => {
        it("returns an empty mutable array if the class has no annotations", () => {
            class Foo {
            }
            const actual = Metadata.getOrCreateAnnotationParamTypes(Foo);
            expect(actual).not.equal(emptyArray);
            expect(actual).length(0);
        });
        it("returns added annotations", () => {
            class Foo {
            }
            const a = Metadata.getOrCreateAnnotationParamTypes(Foo);
            a.push("test");
            const actual = Metadata.getOrCreateAnnotationParamTypes(Foo);
            expect(actual).length(1);
            expect(actual[0]).equal("test");
        });
    });
});
