var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __metadata = (this && this.__metadata) || function (k, v) {
    if (typeof Reflect === "object" && typeof Reflect.metadata === "function") return Reflect.metadata(k, v);
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import { expect } from "chai";
import { attr } from "../components/attributes.js";
import { customElement, FASTElement } from "../components/fast-element.js";
import { observable } from "../observation/observable.js";
import { Updates } from "../observation/update-queue.js";
import { fixture, uniqueElementName } from "../testing/fixture.js";
import { css } from "./css.js";
describe("CSSBindingDirective", () => {
    const name = uniqueElementName();
    const styles = css `.foo { color: ${x => x.color} } .bar { height: ${x => x.size} }`;
    const cssVar1 = styles.behaviors[0].targetAspect;
    const cssVar2 = styles.behaviors[1].targetAspect;
    let TestComponent = class TestComponent extends FASTElement {
        constructor() {
            super(...arguments);
            this.color = "red";
            this.size = "300px";
        }
    };
    __decorate([
        observable,
        __metadata("design:type", String)
    ], TestComponent.prototype, "color", void 0);
    __decorate([
        attr,
        __metadata("design:type", Object)
    ], TestComponent.prototype, "size", void 0);
    TestComponent = __decorate([
        customElement({
            name,
            styles
        })
    ], TestComponent);
    it("sets the model's value to the specified property on the host", () => __awaiter(void 0, void 0, void 0, function* () {
        const { connect, disconnect, element } = yield fixture(name);
        yield connect();
        expect(element.style.getPropertyValue(cssVar1)).equals("red");
        expect(element.style.getPropertyValue(cssVar2)).equals("300px");
        yield disconnect();
    }));
    it("updates the specified property on the host when the model value changes", () => __awaiter(void 0, void 0, void 0, function* () {
        const { connect, disconnect, element } = yield fixture(name);
        yield connect();
        element.color = "blue";
        yield Updates.next();
        expect(element.style.getPropertyValue(cssVar1)).equals("blue");
        expect(element.style.getPropertyValue(cssVar2)).equals("300px");
        element.size = "400px";
        yield Updates.next();
        expect(element.style.getPropertyValue(cssVar1)).equals("blue");
        expect(element.style.getPropertyValue(cssVar2)).equals("400px");
        element.color = "red";
        element.size = "300px";
        yield Updates.next();
        expect(element.style.getPropertyValue(cssVar1)).equals("red");
        expect(element.style.getPropertyValue(cssVar2)).equals("300px");
        yield disconnect();
    }));
    it("updates the specified property on the host when the styles change via setAttribute", () => __awaiter(void 0, void 0, void 0, function* () {
        const { connect, disconnect, element } = yield fixture(name);
        yield connect();
        element.color = "blue";
        element.size = "400px";
        yield Updates.next();
        expect(element.style.getPropertyValue(cssVar1)).equals("blue");
        expect(element.style.getPropertyValue(cssVar2)).equals("400px");
        element.setAttribute("style", "background: red");
        expect(element.style.getPropertyValue(cssVar1)).equals("blue");
        expect(element.style.getPropertyValue(cssVar2)).equals("400px");
        expect(element.style.getPropertyValue("background")).equals("red");
        yield disconnect();
    }));
});
