var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __metadata = (this && this.__metadata) || function (k, v) {
    if (typeof Reflect === "object" && typeof Reflect.metadata === "function") return Reflect.metadata(k, v);
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import { expect } from "chai";
import { HTMLBindingDirective } from "./html-binding-directive.js";
import { observable } from "../observation/observable.js";
import { html } from "./template.js";
import { createTrackableDOMPolicy, toHTML } from "../__test__/helpers.js";
import { HTMLView } from "./view.js";
import { Updates } from "../observation/update-queue.js";
import { DOM } from "../dom.js";
import { Signal, signal } from "../binding/signal.js";
import { twoWay } from "../binding/two-way.js";
import { Fake } from "../testing/fakes.js";
import { HTMLDirective } from "./html-directive.js";
import { nextId } from "./markup.js";
import { oneWay, listener } from "../binding/one-way.js";
import { oneTime } from "../binding/one-time.js";
describe("The HTML binding directive", () => {
    class Model {
        constructor(value) {
            this.value = null;
            this.trigger = 0;
            this.knownValue = "value";
            this.actionInvokeCount = 0;
            this.value = value;
        }
        forceComputedUpdate() {
            this.trigger++;
        }
        invokeAction() {
            this.actionInvokeCount++;
        }
        get computedValue() {
            const trigger = this.trigger;
            return this.value;
        }
    }
    __decorate([
        observable,
        __metadata("design:type", Object)
    ], Model.prototype, "value", void 0);
    __decorate([
        observable,
        __metadata("design:type", Object)
    ], Model.prototype, "trigger", void 0);
    __decorate([
        observable,
        __metadata("design:type", Object)
    ], Model.prototype, "knownValue", void 0);
    function contentBinding(propertyName = "value") {
        const directive = new HTMLBindingDirective(oneWay(x => x[propertyName]));
        directive.id = nextId();
        directive.targetNodeId = 'r';
        directive.targetTagName = null;
        directive.policy = DOM.policy;
        const node = document.createTextNode(" ");
        const targets = { r: node };
        const behavior = directive.createBehavior();
        const parentNode = document.createElement("div");
        parentNode.appendChild(node);
        return { directive, behavior, node, parentNode, targets };
    }
    function compileDirective(directive, sourceAspect, node) {
        var _a;
        if (sourceAspect) {
            HTMLDirective.assignAspect(directive, sourceAspect);
        }
        if (!node) {
            node = document.createElement("div");
        }
        directive.id = nextId();
        directive.targetNodeId = 'r';
        directive.targetTagName = (_a = node.tagName) !== null && _a !== void 0 ? _a : null;
        directive.policy = DOM.policy;
        const targets = { r: node };
        const behavior = directive.createBehavior();
        const parentNode = document.createElement("div");
        parentNode.appendChild(node);
        return { directive, behavior, node, parentNode, targets };
    }
    function defaultBinding(sourceAspect, policy) {
        const directive = new HTMLBindingDirective(oneWay(x => x.value, policy));
        return compileDirective(directive, sourceAspect);
    }
    function oneTimeBinding(sourceAspect, policy) {
        const directive = new HTMLBindingDirective(oneTime(x => x.value, policy));
        return compileDirective(directive, sourceAspect);
    }
    function signalBinding(signalName, sourceAspect, policy) {
        const directive = new HTMLBindingDirective(signal(x => x.value, signalName, policy));
        return compileDirective(directive, sourceAspect);
    }
    function twoWayBinding(options, sourceAspect, policy) {
        const directive = new HTMLBindingDirective(twoWay(x => x.value, options, policy));
        return compileDirective(directive, sourceAspect);
    }
    function eventBinding(options, sourceAspect) {
        const directive = new HTMLBindingDirective(listener(x => x.invokeAction(), options));
        return compileDirective(directive, sourceAspect);
    }
    context("when binding text content", () => {
        it("initially sets the text of a node", () => {
            const { behavior, node, targets } = contentBinding();
            const model = new Model("This is a test");
            const controller = Fake.viewController(targets, behavior);
            controller.bind(model);
            expect(node.textContent).to.equal(model.value);
        });
        it("updates the text of a node when the expression changes", () => __awaiter(void 0, void 0, void 0, function* () {
            const { behavior, node, targets } = contentBinding();
            const model = new Model("This is a test");
            const controller = Fake.viewController(targets, behavior);
            controller.bind(model);
            expect(node.textContent).to.equal(model.value);
            model.value = "This is another test, different from the first.";
            yield Updates.next();
            expect(node.textContent).to.equal(model.value);
        }));
        it("should not throw if DOM stringified", () => {
            const { behavior, node, targets } = contentBinding();
            const model = new Model("This is a test");
            const controller = Fake.viewController(targets, behavior);
            controller.bind(model);
            expect(node.textContent).to.equal(model.value);
            expect(() => {
                JSON.stringify(node);
            }).to.not.throw();
        });
    });
    context("when binding template content", () => {
        it("initially inserts a view based on the template", () => {
            const { behavior, parentNode, targets } = contentBinding();
            const template = html `This is a template. ${x => x.knownValue}`;
            const model = new Model(template);
            const controller = Fake.viewController(targets, behavior);
            controller.bind(model);
            expect(toHTML(parentNode)).to.equal(`This is a template. value`);
        });
        it("removes an inserted view when the value changes to plain text", () => __awaiter(void 0, void 0, void 0, function* () {
            const { behavior, parentNode, targets } = contentBinding();
            const template = html `This is a template. ${x => x.knownValue}`;
            const model = new Model(template);
            const controller = Fake.viewController(targets, behavior);
            controller.bind(model);
            expect(toHTML(parentNode)).to.equal(`This is a template. value`);
            model.value = "This is a test.";
            yield Updates.next();
            expect(toHTML(parentNode)).to.equal(model.value);
        }));
        it("removes an inserted view when the value changes to null", () => __awaiter(void 0, void 0, void 0, function* () {
            const { behavior, parentNode, targets } = contentBinding();
            const template = html `This is a template. ${x => x.knownValue}`;
            const model = new Model(template);
            const controller = Fake.viewController(targets, behavior);
            controller.bind(model);
            expect(toHTML(parentNode)).to.equal(`This is a template. value`);
            model.value = null;
            yield Updates.next();
            expect(toHTML(parentNode)).to.equal("");
        }));
        it("removes an inserted view when the value changes to undefined", () => __awaiter(void 0, void 0, void 0, function* () {
            const { behavior, parentNode, targets } = contentBinding();
            const template = html `This is a template. ${x => x.knownValue}`;
            const model = new Model(template);
            const controller = Fake.viewController(targets, behavior);
            controller.bind(model);
            expect(toHTML(parentNode)).to.equal(`This is a template. value`);
            model.value = void 0;
            yield Updates.next();
            expect(toHTML(parentNode)).to.equal("");
        }));
        it("updates an inserted view when the value changes to a new template", () => __awaiter(void 0, void 0, void 0, function* () {
            const { behavior, parentNode, targets } = contentBinding();
            const template = html `This is a template. ${x => x.knownValue}`;
            const model = new Model(template);
            const controller = Fake.viewController(targets, behavior);
            controller.bind(model);
            expect(toHTML(parentNode)).to.equal(`This is a template. value`);
            const newTemplate = html `This is a new template ${x => x.knownValue}`;
            model.value = newTemplate;
            yield Updates.next();
            expect(toHTML(parentNode)).to.equal(`This is a new template value`);
        }));
        it("reuses a previous view when the value changes back from a string", () => __awaiter(void 0, void 0, void 0, function* () {
            const { behavior, parentNode, node, targets } = contentBinding();
            const template = html `This is a template. ${x => x.knownValue}`;
            const model = new Model(template);
            const controller = Fake.viewController(targets, behavior);
            controller.bind(model);
            const view = node.$fastView;
            const capturedTemplate = node.$fastTemplate;
            expect(view).to.be.instanceOf(HTMLView);
            expect(capturedTemplate).to.equal(template);
            expect(toHTML(parentNode)).to.equal(`This is a template. value`);
            model.value = "This is a test string.";
            yield Updates.next();
            expect(toHTML(parentNode)).to.equal(model.value);
            model.value = template;
            yield Updates.next();
            const newView = node.$fastView;
            const newCapturedTemplate = node.$fastTemplate;
            expect(newView).to.equal(view);
            expect(newCapturedTemplate).to.equal(capturedTemplate);
            expect(toHTML(parentNode)).to.equal(`This is a template. value`);
        }));
        it("doesn't compose an already composed view", () => __awaiter(void 0, void 0, void 0, function* () {
            const { behavior, parentNode, targets } = contentBinding("computedValue");
            const template = html `This is a template. ${x => x.knownValue}`;
            const model = new Model(template);
            const controller = Fake.viewController(targets, behavior);
            controller.bind(model);
            expect(toHTML(parentNode)).to.equal(`This is a template. value`);
            model.value = template;
            model.forceComputedUpdate();
            yield Updates.next();
            expect(toHTML(parentNode)).to.equal(`This is a template. value`);
        }));
        it("pipes the existing execution context through to the new view", () => {
            const { behavior, parentNode, targets } = contentBinding("computedValue");
            const template = html `This is a template. ${(x, c) => c.parent.testProp}`;
            const model = new Model(template);
            const controller = Fake.viewController(targets, behavior);
            const context = Fake.executionContext();
            context.parent = { testProp: "testing..." };
            controller.bind(model, context);
            expect(toHTML(parentNode)).to.equal(`This is a template. testing...`);
        });
        it("allows interpolated HTML tags in templates using dangerousHTML", () => __awaiter(void 0, void 0, void 0, function* () {
            const { behavior, parentNode, targets } = contentBinding();
            const template = html `${x => html `<${html.partial(x.knownValue)}>Hi there!</${html.partial(x.knownValue)}>`}`;
            const model = new Model(template);
            model.knownValue = "button";
            const controller = Fake.viewController(targets, behavior);
            controller.bind(model);
            expect(toHTML(parentNode)).to.equal(`<button>Hi there!</button>`);
            model.knownValue = "a";
            yield Updates.next();
            expect(toHTML(parentNode)).to.equal(`<a>Hi there!</a>`);
        }));
        it("target node should not stringify $fastView or $fastTemplate", () => {
            const { behavior, node, targets } = contentBinding();
            const template = html `This is a template. ${x => x.knownValue}`;
            const model = new Model(template);
            const controller = Fake.viewController(targets, behavior);
            controller.bind(model);
            const clone = JSON.parse(JSON.stringify(node));
            expect("$fastView" in clone).to.be.false;
            expect("$fastTemplate" in clone).to.be.false;
        });
    });
    context("when unbinding template content", () => {
        it("unbinds a composed view", () => {
            const { behavior, node, parentNode, targets } = contentBinding();
            const template = html `This is a template. ${x => x.knownValue}`;
            const model = new Model(template);
            const controller = Fake.viewController(targets, behavior);
            controller.bind(model);
            const newView = node.$fastView;
            expect(newView.source).to.equal(model);
            expect(toHTML(parentNode)).to.equal(`This is a template. value`);
            controller.unbind();
            expect(newView.source).to.equal(null);
        });
        it("rebinds a previously unbound composed view", () => {
            const { behavior, node, parentNode, targets } = contentBinding();
            const template = html `This is a template. ${x => x.knownValue}`;
            const model = new Model(template);
            const controller = Fake.viewController(targets, behavior);
            controller.bind(model);
            const view = node.$fastView;
            expect(view.source).to.equal(model);
            expect(toHTML(parentNode)).to.equal(`This is a template. value`);
            controller.unbind();
            expect(view.source).to.equal(null);
            controller.bind(model);
            const newView = node.$fastView;
            expect(newView.source).to.equal(model);
            expect(toHTML(parentNode)).to.equal(`This is a template. value`);
        });
    });
    const aspectScenarios = [
        {
            name: "content",
            sourceAspect: "",
            originalValue: "This is a test",
            newValue: "This is another test",
            getValue(node) {
                return node.textContent;
            },
            setValue(node, value) {
                node.textContent = value;
            }
        },
        {
            name: "attribute",
            sourceAspect: "test-attribute",
            originalValue: "This is a test",
            newValue: "This is another test",
            getValue(node) {
                return node.getAttribute("test-attribute");
            },
            setValue(node, value) {
                DOM.setAttribute(node, "test-attribute", value);
            }
        },
        {
            name: "boolean attribute",
            sourceAspect: "?test-boolean-attribute",
            originalValue: true,
            newValue: false,
            getValue(node) {
                return node.hasAttribute("test-boolean-attribute");
            },
            setValue(node, value) {
                DOM.setBooleanAttribute(node, "test-boolean-attribute", value);
            }
        },
        {
            name: "property",
            sourceAspect: ":testProperty",
            originalValue: "This is a test",
            newValue: "This is another test",
            getValue(node) {
                return node.testProperty;
            },
            setValue(node, value) {
                node.testProperty = value;
            }
        },
    ];
    context("when binding on-change", () => {
        for (const aspectScenario of aspectScenarios) {
            it(`sets the initial value of a ${aspectScenario.name} binding`, () => {
                const { behavior, node, targets } = defaultBinding(aspectScenario.sourceAspect);
                const model = new Model(aspectScenario.originalValue);
                const controller = Fake.viewController(targets, behavior);
                controller.bind(model);
                expect(aspectScenario.getValue(node)).to.equal(model.value);
            });
            it(`updates the ${aspectScenario.name} when the model changes`, () => __awaiter(void 0, void 0, void 0, function* () {
                const { behavior, node, targets } = defaultBinding(aspectScenario.sourceAspect);
                const model = new Model(aspectScenario.originalValue);
                const controller = Fake.viewController(targets, behavior);
                controller.bind(model);
                expect(aspectScenario.getValue(node)).to.equal(model.value);
                model.value = aspectScenario.newValue;
                yield Updates.next();
                expect(aspectScenario.getValue(node)).to.equal(model.value);
            }));
            it(`doesn't update the ${aspectScenario.name} after unbind`, () => __awaiter(void 0, void 0, void 0, function* () {
                const { behavior, node, targets } = defaultBinding(aspectScenario.sourceAspect);
                const model = new Model(aspectScenario.originalValue);
                const controller = Fake.viewController(targets, behavior);
                controller.bind(model);
                expect(aspectScenario.getValue(node)).to.equal(model.value);
                controller.unbind();
                model.value = aspectScenario.newValue;
                yield Updates.next();
                expect(aspectScenario.getValue(node)).to.equal(aspectScenario.originalValue);
            }));
            it(`uses the dom policy when setting a ${aspectScenario.name} binding`, () => {
                const policy = createTrackableDOMPolicy();
                const { behavior, node, targets } = defaultBinding(aspectScenario.sourceAspect, policy);
                const model = new Model(aspectScenario.originalValue);
                const controller = Fake.viewController(targets, behavior);
                controller.bind(model);
                expect(aspectScenario.getValue(node)).to.equal(model.value);
                expect(policy.used).to.be.true;
            });
            it("should not throw if DOM stringified", () => {
                const { behavior, node, targets } = defaultBinding(aspectScenario.sourceAspect);
                const model = new Model(aspectScenario.originalValue);
                const controller = Fake.viewController(targets, behavior);
                controller.bind(model);
                expect(() => {
                    JSON.stringify(node);
                }).to.not.throw();
            });
        }
    });
    context("when binding one-time", () => {
        for (const aspectScenario of aspectScenarios) {
            it(`sets the initial value of a ${aspectScenario.name} binding`, () => {
                const { behavior, node, targets } = oneTimeBinding(aspectScenario.sourceAspect);
                const model = new Model(aspectScenario.originalValue);
                const controller = Fake.viewController(targets, behavior);
                controller.bind(model);
                expect(aspectScenario.getValue(node)).to.equal(model.value);
            });
            it(`does not update the ${aspectScenario.name} after the initial set`, () => __awaiter(void 0, void 0, void 0, function* () {
                const { behavior, node, targets } = oneTimeBinding(aspectScenario.sourceAspect);
                const model = new Model(aspectScenario.originalValue);
                const controller = Fake.viewController(targets, behavior);
                controller.bind(model);
                expect(aspectScenario.getValue(node)).to.equal(aspectScenario.originalValue);
                model.value = aspectScenario.newValue;
                yield Updates.next();
                expect(aspectScenario.getValue(node)).to.equal(aspectScenario.originalValue);
            }));
            it(`doesn't update the ${aspectScenario.name} after unbind`, () => __awaiter(void 0, void 0, void 0, function* () {
                const { behavior, node, targets } = oneTimeBinding(aspectScenario.sourceAspect);
                const model = new Model(aspectScenario.originalValue);
                const controller = Fake.viewController(targets, behavior);
                controller.bind(model);
                expect(aspectScenario.getValue(node)).to.equal(aspectScenario.originalValue);
                controller.unbind();
                model.value = aspectScenario.newValue;
                yield Updates.next();
                expect(aspectScenario.getValue(node)).to.equal(aspectScenario.originalValue);
            }));
            it(`uses the dom policy when setting a ${aspectScenario.name} binding`, () => {
                const policy = createTrackableDOMPolicy();
                const { behavior, node, targets } = oneTimeBinding(aspectScenario.sourceAspect, policy);
                const model = new Model(aspectScenario.originalValue);
                const controller = Fake.viewController(targets, behavior);
                controller.bind(model);
                expect(aspectScenario.getValue(node)).to.equal(model.value);
                expect(policy.used).to.be.true;
            });
            it("should not throw if DOM stringified", () => {
                const { behavior, node, targets } = oneTimeBinding(aspectScenario.sourceAspect);
                const model = new Model(aspectScenario.originalValue);
                const controller = Fake.viewController(targets, behavior);
                controller.bind(model);
                expect(() => {
                    JSON.stringify(node);
                }).to.not.throw();
            });
        }
    });
    context("when binding with a signal", () => {
        for (const aspectScenario of aspectScenarios) {
            it(`sets the initial value of the ${aspectScenario.name} binding`, () => {
                const { behavior, node, targets } = signalBinding("test-signal", aspectScenario.sourceAspect);
                const model = new Model(aspectScenario.originalValue);
                const controller = Fake.viewController(targets, behavior);
                controller.bind(model);
                expect(aspectScenario.getValue(node)).to.equal(model.value);
            });
            it(`updates the ${aspectScenario.name} only when the signal is sent`, () => __awaiter(void 0, void 0, void 0, function* () {
                const signalName = "test-signal";
                const { behavior, node, targets } = signalBinding(signalName, aspectScenario.sourceAspect);
                const model = new Model(aspectScenario.originalValue);
                const controller = Fake.viewController(targets, behavior);
                controller.bind(model);
                expect(aspectScenario.getValue(node)).to.equal(aspectScenario.originalValue);
                model.value = aspectScenario.newValue;
                yield Updates.next();
                expect(aspectScenario.getValue(node)).to.equal(aspectScenario.originalValue);
                Signal.send(signalName);
                yield Updates.next();
                expect(aspectScenario.getValue(node)).to.equal(model.value);
            }));
            it(`doesn't respond to signals for a ${aspectScenario.name} binding after unbind`, () => __awaiter(void 0, void 0, void 0, function* () {
                const signalName = "test-signal";
                const { behavior, node, targets } = signalBinding(signalName, aspectScenario.sourceAspect);
                const model = new Model(aspectScenario.originalValue);
                const controller = Fake.viewController(targets, behavior);
                controller.bind(model);
                expect(aspectScenario.getValue(node)).to.equal(model.value);
                controller.unbind();
                model.value = aspectScenario.newValue;
                Signal.send(signalName);
                yield Updates.next();
                expect(aspectScenario.getValue(node)).to.equal(aspectScenario.originalValue);
            }));
            it(`uses the dom policy when setting a ${aspectScenario.name} binding`, () => {
                const policy = createTrackableDOMPolicy();
                const { behavior, node, targets } = signalBinding("test-signal", aspectScenario.sourceAspect, policy);
                const model = new Model(aspectScenario.originalValue);
                const controller = Fake.viewController(targets, behavior);
                controller.bind(model);
                expect(aspectScenario.getValue(node)).to.equal(model.value);
                expect(policy.used).to.be.true;
            });
            it("should not throw if DOM stringified", () => {
                const { behavior, node, targets } = signalBinding("test-signal", aspectScenario.sourceAspect);
                const model = new Model(aspectScenario.originalValue);
                const controller = Fake.viewController(targets, behavior);
                controller.bind(model);
                expect(() => {
                    JSON.stringify(node);
                }).to.not.throw();
            });
        }
    });
    context("when binding two-way", () => {
        for (const aspectScenario of aspectScenarios) {
            it(`sets the initial value of the ${aspectScenario.name} binding`, () => {
                const { behavior, node, targets } = twoWayBinding({}, aspectScenario.sourceAspect);
                const model = new Model(aspectScenario.originalValue);
                const controller = Fake.viewController(targets, behavior);
                controller.bind(model);
                expect(aspectScenario.getValue(node)).to.equal(model.value);
            });
            it(`updates the ${aspectScenario.name} when the model changes`, () => __awaiter(void 0, void 0, void 0, function* () {
                const { behavior, node, targets } = twoWayBinding({}, aspectScenario.sourceAspect);
                const model = new Model(aspectScenario.originalValue);
                const controller = Fake.viewController(targets, behavior);
                controller.bind(model);
                expect(aspectScenario.getValue(node)).to.equal(aspectScenario.originalValue);
                model.value = aspectScenario.newValue;
                yield Updates.next();
                expect(aspectScenario.getValue(node)).to.equal(model.value);
            }));
            it(`updates the model when a change event fires for the ${aspectScenario.name}`, () => __awaiter(void 0, void 0, void 0, function* () {
                const { behavior, node, targets } = twoWayBinding({}, aspectScenario.sourceAspect);
                const model = new Model(aspectScenario.originalValue);
                const controller = Fake.viewController(targets, behavior);
                controller.bind(model);
                expect(aspectScenario.getValue(node)).to.equal(aspectScenario.originalValue);
                aspectScenario.setValue(node, aspectScenario.newValue);
                node.dispatchEvent(new CustomEvent("change"));
                yield Updates.next();
                expect(model.value).to.equal(aspectScenario.newValue);
            }));
            it(`updates the model when a change event fires for the ${aspectScenario.name} with conversion`, () => __awaiter(void 0, void 0, void 0, function* () {
                const fromView = value => "fixed value";
                const { behavior, node, targets } = twoWayBinding({ fromView }, aspectScenario.sourceAspect);
                const model = new Model(aspectScenario.originalValue);
                const controller = Fake.viewController(targets, behavior);
                controller.bind(model);
                expect(aspectScenario.getValue(node)).to.equal(aspectScenario.originalValue);
                aspectScenario.setValue(node, aspectScenario.newValue);
                node.dispatchEvent(new CustomEvent("change"));
                yield Updates.next();
                expect(model.value).to.equal("fixed value");
            }));
            it(`updates the model when a configured event fires for the ${aspectScenario.name}`, () => __awaiter(void 0, void 0, void 0, function* () {
                const changeEvent = "foo";
                const { behavior, node, targets } = twoWayBinding({ changeEvent }, aspectScenario.sourceAspect);
                const model = new Model(aspectScenario.originalValue);
                const controller = Fake.viewController(targets, behavior);
                controller.bind(model);
                expect(aspectScenario.getValue(node)).to.equal(aspectScenario.originalValue);
                aspectScenario.setValue(node, aspectScenario.newValue);
                node.dispatchEvent(new CustomEvent(changeEvent));
                yield Updates.next();
                expect(model.value).to.equal(aspectScenario.newValue);
            }));
            it(`doesn't update the ${aspectScenario.name} after unbind`, () => __awaiter(void 0, void 0, void 0, function* () {
                const { behavior, node, targets } = twoWayBinding({}, aspectScenario.sourceAspect);
                const model = new Model(aspectScenario.originalValue);
                const controller = Fake.viewController(targets, behavior);
                controller.bind(model);
                expect(aspectScenario.getValue(node)).to.equal(model.value);
                controller.unbind();
                model.value = aspectScenario.newValue;
                yield Updates.next();
                expect(aspectScenario.getValue(node)).to.equal(aspectScenario.originalValue);
            }));
            it(`uses the dom policy when setting a ${aspectScenario.name} binding`, () => {
                const policy = createTrackableDOMPolicy();
                const { behavior, node, targets } = twoWayBinding({}, aspectScenario.sourceAspect, policy);
                const model = new Model(aspectScenario.originalValue);
                const controller = Fake.viewController(targets, behavior);
                controller.bind(model);
                expect(aspectScenario.getValue(node)).to.equal(model.value);
                expect(policy.used).to.be.true;
            });
            it("should not throw if DOM stringified", () => {
                const { behavior, node, targets } = twoWayBinding({}, aspectScenario.sourceAspect);
                const model = new Model(aspectScenario.originalValue);
                const controller = Fake.viewController(targets, behavior);
                controller.bind(model);
                expect(() => {
                    JSON.stringify(node);
                }).to.not.throw();
            });
        }
    });
    context("when binding events", () => {
        it("does not invoke the method on bind", () => {
            const { behavior, targets } = eventBinding({}, "@my-event");
            const model = new Model("Test value.");
            const controller = Fake.viewController(targets, behavior);
            controller.bind(model);
            expect(model.actionInvokeCount).to.equal(0);
        });
        it("invokes the method each time the event is raised", () => {
            const { behavior, node, targets } = eventBinding({}, "@my-event");
            const model = new Model("Test value.");
            const controller = Fake.viewController(targets, behavior);
            controller.bind(model);
            expect(model.actionInvokeCount).to.equal(0);
            node.dispatchEvent(new CustomEvent("my-event"));
            expect(model.actionInvokeCount).to.equal(1);
            node.dispatchEvent(new CustomEvent("my-event"));
            expect(model.actionInvokeCount).to.equal(2);
            node.dispatchEvent(new CustomEvent("my-event"));
            expect(model.actionInvokeCount).to.equal(3);
        });
        it("invokes the method one time for a one time event", () => {
            const { behavior, node, targets } = eventBinding({ once: true }, "@my-event");
            const model = new Model("Test value.");
            const controller = Fake.viewController(targets, behavior);
            controller.bind(model);
            expect(model.actionInvokeCount).to.equal(0);
            node.dispatchEvent(new CustomEvent("my-event"));
            expect(model.actionInvokeCount).to.equal(1);
            node.dispatchEvent(new CustomEvent("my-event"));
            expect(model.actionInvokeCount).to.equal(1);
        });
        it("does not invoke the method when unbound", () => {
            const { behavior, node, targets } = eventBinding({}, "@my-event");
            const model = new Model("Test value.");
            const controller = Fake.viewController(targets, behavior);
            controller.bind(model);
            expect(model.actionInvokeCount).to.equal(0);
            node.dispatchEvent(new CustomEvent("my-event"));
            expect(model.actionInvokeCount).to.equal(1);
            controller.unbind();
            node.dispatchEvent(new CustomEvent("my-event"));
            expect(model.actionInvokeCount).to.equal(1);
        });
        it("should not throw if DOM stringified", () => {
            const { behavior, targets, node } = eventBinding({}, "@my-event");
            const model = new Model("Test value.");
            const controller = Fake.viewController(targets, behavior);
            controller.bind(model);
            expect(() => {
                JSON.stringify(node);
            }).to.not.throw();
        });
    });
    context('when binding classList', () => {
        function updateTarget(target, directive, value) {
            directive.updateTarget(target, directive.targetAspect, value, Fake.viewController());
        }
        function createClassBinding(element) {
            const directive = new HTMLBindingDirective(oneWay(() => ""));
            return compileDirective(directive, ":classList", element);
        }
        it('adds and removes own classes', () => {
            const element = document.createElement("div");
            element.classList.add("foo");
            element.classList.add("bar");
            const { directive: observerA } = createClassBinding(element);
            const { directive: observerB } = createClassBinding(element);
            const contains = element.classList.contains.bind(element.classList);
            expect(contains('foo') && contains('bar')).true;
            updateTarget(element, observerA, ' xxx \t\r\n\v\f yyy  ');
            expect(contains('foo') && contains('bar')).true;
            expect(contains('xxx') && contains('yyy')).true;
            updateTarget(element, observerA, '');
            expect(contains('foo') && contains('bar')).true;
            expect(contains('xxx') || contains('yyy')).false;
            updateTarget(element, observerB, 'bbb');
            expect(contains('foo') && contains('bar')).true;
            expect(contains('bbb')).true;
            updateTarget(element, observerB, 'aaa');
            expect(contains('foo') && contains('bar')).true;
            expect(contains('aaa') && !contains('bbb')).true;
            updateTarget(element, observerA, 'foo bar');
            expect(contains('foo') && contains('bar')).true;
            updateTarget(element, observerA, '');
            expect(contains('foo') || contains('bar')).false;
            updateTarget(element, observerA, 'foo');
            expect(contains('foo')).true;
            updateTarget(element, observerA, null);
            expect(contains('foo')).false;
            updateTarget(element, observerA, 'foo');
            expect(contains('foo')).true;
            updateTarget(element, observerA, undefined);
            expect(contains('foo')).false;
        });
        it("should not throw if DOM stringified", () => {
            const directive = new HTMLBindingDirective(oneWay(() => ""));
            const { behavior, node, targets } = compileDirective(directive, ":classList");
            HTMLDirective.assignAspect(directive, ":classList");
            const model = new Model("Test value.");
            const controller = Fake.viewController(targets, behavior);
            controller.bind(model);
            expect(() => {
                JSON.stringify(node);
            }).to.not.throw();
        });
    });
});
