var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __metadata = (this && this.__metadata) || function (k, v) {
    if (typeof Reflect === "object" && typeof Reflect.metadata === "function") return Reflect.metadata(k, v);
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import { expect } from "chai";
import { children, ChildrenDirective } from "./children.js";
import { observable } from "../observation/observable.js";
import { elements } from "./node-observation.js";
import { Updates } from "../observation/update-queue.js";
import { Fake } from "../testing/fakes.js";
import { html } from "./template.js";
import { ref } from "./ref.js";
import "../state/state.js";
describe("The children", () => {
    context("template function", () => {
        it("returns an ChildrenDirective", () => {
            const directive = children("test");
            expect(directive).to.be.instanceOf(ChildrenDirective);
        });
    });
    context("directive", () => {
        it("creates a behavior by returning itself", () => {
            const directive = children("test");
            const behavior = directive.createBehavior();
            expect(behavior).to.equal(behavior);
        });
    });
    context("behavior", () => {
        class Model {
        }
        __decorate([
            observable,
            __metadata("design:type", Object)
        ], Model.prototype, "nodes", void 0);
        function createAndAppendChildren(host, elementName = "div") {
            const children = new Array(10);
            for (let i = 0, ii = children.length; i < ii; ++i) {
                const child = document.createElement(i % 1 === 0 ? elementName : "div");
                children[i] = child;
                host.appendChild(child);
            }
            return children;
        }
        function createDOM(elementName = "div") {
            const host = document.createElement("div");
            const children = createAndAppendChildren(host, elementName);
            const nodeId = 'r';
            const targets = { [nodeId]: host };
            return { host, children, targets, nodeId };
        }
        it("gathers child nodes", () => {
            const { host, children, targets, nodeId } = createDOM();
            const behavior = new ChildrenDirective({
                property: "nodes",
            });
            behavior.targetNodeId = nodeId;
            const model = new Model();
            const controller = Fake.viewController(targets, behavior);
            controller.bind(model);
            expect(model.nodes).members(children);
        });
        it("gathers child nodes with a filter", () => {
            const { host, children, targets, nodeId } = createDOM("foo-bar");
            const behavior = new ChildrenDirective({
                property: "nodes",
                filter: elements("foo-bar"),
            });
            behavior.targetNodeId = nodeId;
            const model = new Model();
            const controller = Fake.viewController(targets, behavior);
            controller.bind(model);
            expect(model.nodes).members(children.filter(elements("foo-bar")));
        });
        it("updates child nodes when they change", () => __awaiter(void 0, void 0, void 0, function* () {
            const { host, children, targets, nodeId } = createDOM("foo-bar");
            const behavior = new ChildrenDirective({
                property: "nodes",
            });
            behavior.targetNodeId = nodeId;
            const model = new Model();
            const controller = Fake.viewController(targets, behavior);
            controller.bind(model);
            expect(model.nodes).members(children);
            const updatedChildren = children.concat(createAndAppendChildren(host));
            yield Updates.next();
            expect(model.nodes).members(updatedChildren);
        }));
        it("updates child nodes when they change with a filter", () => __awaiter(void 0, void 0, void 0, function* () {
            const { host, children, targets, nodeId } = createDOM("foo-bar");
            const behavior = new ChildrenDirective({
                property: "nodes",
                filter: elements("foo-bar"),
            });
            behavior.targetNodeId = nodeId;
            const model = new Model();
            const controller = Fake.viewController(targets, behavior);
            controller.bind(model);
            expect(model.nodes).members(children);
            const updatedChildren = children.concat(createAndAppendChildren(host));
            yield Updates.next();
            expect(model.nodes).members(updatedChildren.filter(elements("foo-bar")));
        }));
        it("updates subtree nodes when they change with a selector", () => __awaiter(void 0, void 0, void 0, function* () {
            const { host, children, targets, nodeId } = createDOM("foo-bar");
            const subtreeElement = "foo-bar-baz";
            const subtreeChildren = [];
            for (let child of children) {
                for (let i = 0; i < 3; ++i) {
                    const subChild = document.createElement("foo-bar-baz");
                    subtreeChildren.push(subChild);
                    child.appendChild(subChild);
                }
            }
            const behavior = new ChildrenDirective({
                property: "nodes",
                subtree: true,
                selector: subtreeElement,
            });
            behavior.targetNodeId = nodeId;
            const model = new Model();
            const controller = Fake.viewController(targets, behavior);
            controller.bind(model);
            expect(model.nodes).members(subtreeChildren);
            const newChildren = createAndAppendChildren(host);
            for (let child of newChildren) {
                for (let i = 0; i < 3; ++i) {
                    const subChild = document.createElement("foo-bar-baz");
                    subtreeChildren.push(subChild);
                    child.appendChild(subChild);
                }
            }
            yield Updates.next();
            expect(model.nodes).members(subtreeChildren);
        }));
        it("clears and unwatches when unbound", () => __awaiter(void 0, void 0, void 0, function* () {
            const { host, children, targets, nodeId } = createDOM("foo-bar");
            const behavior = new ChildrenDirective({
                property: "nodes",
            });
            behavior.targetNodeId = nodeId;
            const model = new Model();
            const controller = Fake.viewController(targets, behavior);
            controller.bind(model);
            expect(model.nodes).members(children);
            behavior.unbind(controller);
            expect(model.nodes).members([]);
            host.appendChild(document.createElement("div"));
            yield Updates.next();
            expect(model.nodes).members([]);
        }));
        it("re-watches when re-bound", () => __awaiter(void 0, void 0, void 0, function* () {
            const { host, children, targets, nodeId } = createDOM("foo-bar");
            const behavior = new ChildrenDirective({
                property: "nodes",
            });
            behavior.targetNodeId = nodeId;
            const model = new Model();
            const controller = Fake.viewController(targets, behavior);
            controller.bind(model);
            behavior.unbind(controller);
            behavior.bind(controller);
            const element = document.createElement("div");
            host.appendChild(element);
            yield Updates.next();
            expect(model.nodes.includes(element)).to.equal(true);
        }));
        it("should not throw if DOM stringified", () => {
            const template = html `
                <div id="test"
                     ${children("nodes")}
                     ${ref("reference")}>
                </div>
            `;
            const view = template.create();
            const model = new Model();
            view.bind(model);
            expect(() => {
                JSON.stringify(model.reference);
            }).to.not.throw();
            view.unbind();
        });
        it("supports multiple directives for the same element", () => __awaiter(void 0, void 0, void 0, function* () {
            const { host, targets, nodeId } = createDOM("foo-bar");
            class MultipleDirectivesModel {
                constructor() {
                    this.elements = [];
                    this.text = [];
                }
            }
            __decorate([
                observable,
                __metadata("design:type", Array)
            ], MultipleDirectivesModel.prototype, "elements", void 0);
            __decorate([
                observable,
                __metadata("design:type", Array)
            ], MultipleDirectivesModel.prototype, "text", void 0);
            const elementsDirective = new ChildrenDirective({
                property: "elements",
                filter: elements(),
            });
            const textDirective = new ChildrenDirective({
                property: "text",
                filter: (value) => value.nodeType === Node.TEXT_NODE,
            });
            elementsDirective.targetNodeId = nodeId;
            textDirective.targetNodeId = nodeId;
            const model = new MultipleDirectivesModel();
            const controller = Fake.viewController(targets, elementsDirective, textDirective);
            controller.bind(model);
            elementsDirective.bind(controller);
            textDirective.bind(controller);
            const element = document.createElement("div");
            const text = document.createTextNode("text");
            host.appendChild(element);
            host.appendChild(text);
            yield Updates.next();
            expect(model.elements.includes(element)).to.equal(true);
            expect(model.elements.includes(text)).to.equal(false);
            expect(model.text.includes(text)).to.equal(true);
            expect(model.text.includes(element)).to.equal(false);
        }));
    });
});
