var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __metadata = (this && this.__metadata) || function (k, v) {
    if (typeof Reflect === "object" && typeof Reflect.metadata === "function") return Reflect.metadata(k, v);
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import { expect } from "chai";
import { slotted, SlottedDirective } from "./slotted.js";
import { ref } from "./ref.js";
import { observable } from "../observation/observable.js";
import { elements } from "./node-observation.js";
import { Updates } from "../observation/update-queue.js";
import { Fake } from "../testing/fakes.js";
import { html } from "./template.js";
describe("The slotted", () => {
    context("template function", () => {
        it("returns an ChildrenDirective", () => {
            const directive = slotted("test");
            expect(directive).to.be.instanceOf(SlottedDirective);
        });
    });
    context("directive", () => {
        it("creates a behavior by returning itself", () => {
            const nodeId = 'r';
            const directive = slotted("test");
            directive.targetNodeId = nodeId;
            const behavior = directive.createBehavior();
            expect(behavior).to.equal(directive);
        });
    });
    context("behavior", () => {
        class Model {
        }
        __decorate([
            observable,
            __metadata("design:type", Object)
        ], Model.prototype, "nodes", void 0);
        function createAndAppendChildren(host, elementName = "div") {
            const children = new Array(10);
            for (let i = 0, ii = children.length; i < ii; ++i) {
                const child = document.createElement(i % 1 === 0 ? elementName : "div");
                children[i] = child;
                host.appendChild(child);
            }
            return children;
        }
        function createDOM(elementName = "div") {
            const host = document.createElement("div");
            const slot = document.createElement("slot");
            const shadowRoot = host.attachShadow({ mode: "open" });
            const children = createAndAppendChildren(host, elementName);
            const nodeId = 'r';
            const targets = { [nodeId]: slot };
            shadowRoot.appendChild(slot);
            return { host, slot, children, targets, nodeId };
        }
        function createController(source, targets) {
            return {
                source,
                targets,
                context: Fake.executionContext(),
                isBound: false,
                onUnbind() {
                }
            };
        }
        it("gathers nodes from a slot", () => {
            const { children, targets, nodeId } = createDOM();
            const behavior = new SlottedDirective({ property: "nodes" });
            behavior.targetNodeId = nodeId;
            const model = new Model();
            const controller = createController(model, targets);
            behavior.bind(controller);
            expect(model.nodes).members(children);
        });
        it("gathers nodes from a slot with a filter", () => {
            const { targets, nodeId, children } = createDOM("foo-bar");
            const behavior = new SlottedDirective({
                property: "nodes",
                filter: elements("foo-bar"),
            });
            behavior.targetNodeId = nodeId;
            const model = new Model();
            const controller = createController(model, targets);
            behavior.bind(controller);
            expect(model.nodes).members(children.filter(elements("foo-bar")));
        });
        it("updates when slotted nodes change", () => __awaiter(void 0, void 0, void 0, function* () {
            const { host, slot, children, targets, nodeId } = createDOM("foo-bar");
            const behavior = new SlottedDirective({ property: "nodes" });
            behavior.targetNodeId = nodeId;
            const model = new Model();
            const controller = createController(model, targets);
            behavior.bind(controller);
            expect(model.nodes).members(children);
            const updatedChildren = children.concat(createAndAppendChildren(host));
            yield Updates.next();
            expect(model.nodes).members(updatedChildren);
        }));
        it("updates when slotted nodes change with a filter", () => __awaiter(void 0, void 0, void 0, function* () {
            const { host, slot, children, targets, nodeId } = createDOM("foo-bar");
            const behavior = new SlottedDirective({
                property: "nodes",
                filter: elements("foo-bar"),
            });
            behavior.targetNodeId = nodeId;
            const model = new Model();
            const controller = createController(model, targets);
            behavior.bind(controller);
            expect(model.nodes).members(children);
            const updatedChildren = children.concat(createAndAppendChildren(host));
            yield Updates.next();
            expect(model.nodes).members(updatedChildren.filter(elements("foo-bar")));
        }));
        it("clears and unwatches when unbound", () => __awaiter(void 0, void 0, void 0, function* () {
            const { host, slot, children, targets, nodeId } = createDOM("foo-bar");
            const behavior = new SlottedDirective({ property: "nodes" });
            behavior.targetNodeId = nodeId;
            const model = new Model();
            const controller = createController(model, targets);
            behavior.bind(controller);
            expect(model.nodes).members(children);
            behavior.unbind(controller);
            expect(model.nodes).members([]);
            host.appendChild(document.createElement("div"));
            yield Updates.next();
            expect(model.nodes).members([]);
        }));
        it("should not throw if DOM stringified", () => {
            const template = html `
                <slot id="test"
                     ${slotted("nodes")}
                     ${ref("reference")}>
                </div>
            `;
            const view = template.create();
            const model = new Model();
            view.bind(model);
            expect(() => {
                JSON.stringify(model.reference);
            }).to.not.throw();
            view.unbind();
        });
    });
});
